package rvl.swv;
import rvl.swv.*;
import java.util.*;
import java.util.regex.*;
import java.io.*;

/** <TT>StatWeave</TT> is the main class for weaving and tangling source files.
It constructs an instance of {@link FileInterface} for the required file type, plus
one or more instances of
{@link EngineInterface} as needed based on the statistical engine(s) required
to process the source file.
<p>
The {@link #main} method provides a command-line interface that reads options
and the source filename from the command line.  The main work is done by
{@link #weave}, which is called by {@link #main} but could also be called
by, say, a GUI.
<p>
To use this class, we also require that a configuration file be defined; its
default name is <TT>statweave.cfg</TT> on Windows systems, and <TT>.statweave</TT>
on all other platforms.  <TT>StatWeave</TT> looks for this file in the home directory.

*/
public class StatWeave {

    String fileName, fileDir, baseName, target;

    String fmt=null, syn=null;  // Other class could set these to non-null to override default detection methods

    int tagNo, chunkNo, mopup, recIndex = 1;

    boolean weaving,       // Flag: true if weaving, false if tangling
        dryrun;            // True is equivalent to a universal eval=FALSE

    FileInterface filei;   // Handler for reading source/writing result
    SyntaxInterface syni;  // Handler for syntax.
    HashMap<String, EngineInterface> engines;  // active engines
    Vector<EngineInterface>  allEngines;  // all engines, in order of creating (and execution)
    Vector<FigFile> allFigs;  // all figures
    HashMap<String, String> engineMap;  // maps languages to engine names
    HashMap<String, String> labeledChunks; // This is where labeled code chunks are stored
    HashMap<String, FigFile> figFiles;         // Map for matching figures w/ signatures

    HashMap<String,String> recallCode,
        recallCodeSig, recallOutSig, recallFigSig;  // Maps for recall constructs
        // all 4 have label as the key
        // recallCode has the actual code listing as value
        // recall CodeSig, recallFigSig, and recallOutSig have the signatures as values

    /** Option lists.  Via generics, we'll force <TT>String</TT>s only for both keys and values */
    HashMap<String,String> factoryOpts, globalOpts, currentOpts;

    /** <TT>HashMap</TT> of individual <TT>HashMaps</TT> to use for language-specific options */
    HashMap<String, HashMap<String,String>> langSpecOpts;

    /** Map chunk signatures to options in effect for that chunk for later recall */
    HashMap<String, HashMap<String,String>> optHistory;

    /** Configuration information */
    Properties config;

    public PrintStream errStream = System.err;     // Where to direct error messages, e.g. stack traces

    boolean isWindows;   // If true, we'll be case-insensitive at certain times


  /** Constructor */
    public StatWeave(String cfgFile) {
        factoryOpts = new HashMap<String,String>(20);
        globalOpts = new HashMap<String,String>(20);
        currentOpts = new HashMap<String,String>(20);
        langSpecOpts = new HashMap<String,HashMap<String,String>>(4);
        optHistory = new HashMap<String,HashMap<String,String>>();
        engineMap = new HashMap<String,String>(4);
        engines = new HashMap<String, EngineInterface>(4);
        allEngines = new Vector<EngineInterface>();
        allFigs = new Vector<FigFile>();
        figFiles = new HashMap<String, FigFile>();
        labeledChunks = new HashMap<String, String>(4);
        recallCode = new HashMap<String,String>();
        recallCodeSig = new HashMap<String,String>();
        recallOutSig = new HashMap<String,String>();
        recallFigSig = new HashMap<String,String>();
        readConfig(cfgFile);
        setDefaults();
    }

  /** Initialize options */
    void setDefaults() {
        String fopts = "echo, eval";  // -- not needed !fig, !hide, ompt=\"> \", !showref,";
        parseOptions(fopts, factoryOpts);
        parseOptions(getConfig("Global.options"), factoryOpts);
    }

  /** Display a non-error message */
    public static void message(String msg) {
        System.out.println(msg);
    }

  /** Called by interface classes to report a warning */
    public static void warning (String msg) {
        System.err.println("WARNING -- " + msg);
    }

  /**
    Called by interface classes to report an error.
    A recommended exit code is also provided
  */
    public void error (String msg, int errCode) {
        System.err.println("ERROR -- " + msg);
        System.exit(errCode);
    }

  /** Return the specified option from <TT>currentOpts</TT> */
    public String getOption(String optName) {
        return currentOpts.get(optName);
    }
  /** Return the specified option from <TT>currentOpts</TT>.
      If it is null, return the value of <tt>dfault</tt>
  */
    public String getOption(String optName, String dfault) {
        String opt = currentOpts.get(optName);
        if (opt == null) opt = dfault;
        return opt;
    }

  /** Returns <TT>true</TT> if option named <TT>optname</TT> starts with <TT>T</TT> */
    public boolean isTrue(String optName) {
        String o = getOption(optName);
        if (o==null) return false;
        return o.startsWith("T");
    }

  /** Returns a dimension in inches for named option.
    Option may end in "in", "cm", "mm", "pt", "px", or nothing.
    If nothing, we guess the units based on the value.
    Note: px units are considered same as pt
    If option doesn't exist, <tt>dfault</tt> is returned
  */
    public double getDim(String optName, double dfault) {
        double divisor = 0, x;
        String spec = getOption(optName);
        if (spec==null) return dfault;

        String dim = spec.trim();
        if (dim.endsWith("in")) divisor = 1;
        else if(dim.endsWith("cm")) divisor = 2.54;
        else if(dim.endsWith("mm")) divisor = 25.4;
        else if(dim.endsWith("pt")||dim.endsWith("px")) divisor = 72.0;
        if (divisor > 0) dim = dim.substring(0, dim.length()-2);
        try {
            x = new Double(dim).doubleValue();
        }
        catch(NumberFormatException nfe) {
            message("Illegal dimension " + spec + " -- fallback value is 4.0in");
            x = 4.0;
            divisor = 1;
        }
        if (divisor==0) divisor = x < 6.0001 ? 1.0
                                : x < 15.001 ? 2.54
                                : x < 100    ? 25.4
                                :              72.0;
        return x / divisor;
    }


  /** Parse the option list and add/change to the given <TT>HashMap</TT>.<p>
   We extend Sweave's conventions by not requiring an equal sign for boolean options.
   For example, <TT>fig</TT> is equivalent to <TT>fig=TRUE</TT>,
   and <TT>!hide</TT> is equivalent to <TT>hide=FALSE</TT>.
  */
// New version handles commas within quotes, allows quotes within quotes
// e.g.: list = "abc, def, ""quoted"", xyz"
void parseOptions(String optlist, HashMap<String,String> map) {
        optlist += ","; // easier to program if there's a delimiter at end
        int len = optlist.length(), p0=0, p1, p2;
        try {
            for (p0=0; p0 < len; ) {
                p1 = optlist.indexOf(",", p0);
                p2 = optlist.indexOf("=", p0);
                if ((p1 > 0) && (p1 < p2 || p2 < 0)) { // we have a comma before any equals
                    String key = optlist.substring(p0, p1).trim();
                    if (key.length() > 0) {
                        if (key.startsWith("!"))
                            map.put(key.substring(1), "FALSE");
                        else if (key.startsWith("-"))
                            map.remove(key.substring(1));
                        else
                            map.put(key, "TRUE");
                    }
                    p0 = p1 + 1; // past 1st comma
                }
                else if(p2 > 0) { // we have an equal sign
                    String key = optlist.substring(p0, p2).trim();
                    for (p0 = p2 + 1; optlist.charAt(p0) == ' '; p0++); // p0 -> 1st nonblank past =
                    if (optlist.charAt(p0) == '"') { // quoted item
                        p1 = p0 + 1;
                        for (boolean pending=true; pending; ) {
                            p2 = optlist.indexOf("\"", p1);
                            if (p2 < 0) throw new NoSuchElementException();
                            if (optlist.charAt(p2+1) == '"') p1 = p2 + 2; // Have double "" -- keep looking
                            else pending = false; // we've got end of quote
                        }
                        String val = optlist.substring(p0,p2+1).trim();
                        val = val.substring(1,val.length()-1);  // strip off quotes
                        val = val.replaceAll("\"\"", "\"");
                        p0 = optlist.indexOf(",", p2) + 1;
                        map.put(key,  val);
                    }
                    else { // unquoted -- look for next entry
                        p1 = optlist.indexOf(",", p0);
                        String val = p1 > 0 ? optlist.substring(p0,p1) : optlist.substring(p0);
                        map.put(key, val.trim());
                        p0 = p1 > 0? p1 + 1 : len;
                    }
                }
                else // nothing more to look at
                    p0 = len;
            }
        }
        catch (NoSuchElementException e) {
            error("Error in option list: "+ optlist.substring(0,p0) + " >>-ERROR->> " + optlist.substring(p0, len-1), 26);
        }
    }


  /** Get configuration information */
    void readConfig(String cfgFile) {
        config = new Properties();
        isWindows = System.getProperty("os.name").startsWith("Win");
        try {
            if (cfgFile == null) {
                String home = System.getProperty("user.home")
                    + System.getProperty("file.separator");
                cfgFile =  isWindows ? home + "statweave.cfg" : home + ".statweave";
            }
            config.load(new FileInputStream(cfgFile));
            message("Loaded configuration information from " + cfgFile);
        }
        catch (FileNotFoundException e) {
            error("Configuration file \"" + cfgFile + "\" not found", 1);
        }
        catch (IOException e) {
            error("I/O error", 2);
        }
    }

  /** accessor to configuration properties */
    public String getConfig(String property) {
        return config.getProperty(property);
    }

  /** return true if weaving, false if tangling */
    public boolean isWeaving() {
        return weaving;
    }

  /** return directory of source file */
    public String getDir() {
        return fileDir;
    }

  /** Remove excess blank lines from output (note it starts with "!") */
    String cleanUp(String txt) {
        txt = txt.replaceAll("[ ]+\\n","\n");  // strip trailing spaces
        txt = txt.replaceAll("^[\\n]+", "");   // strip blank lines at beginning
        txt = txt.replaceAll("[\\n]+$", "\n");  // strip blank lines at end
        //txt = txt.replaceAll("\\n\\n\\n", "\n\n");  // compress consecutive blank lines (loose compression)
        txt = txt.replaceAll("\\n\\n", "\n");  // compress consecutive blank lines (tighter compression)
        return txt;
    }

  /** If name ends with one of the space-delimited strings in list.
      the matching token is returned; otherwise, null.
      This method may be called by a child class when parsing configuration keys
  */
    String match(String name, String list) {
        if (list == null) return null; // safety exit
        StringTokenizer st = new StringTokenizer(list);
        while (st.hasMoreTokens()) {
            String tok = st.nextToken();
            if (name.endsWith(tok))
                return tok;
        }
        return null;
    }

  /** Examine name of source file name to decide format.  Return name of file format
  and set baseName appropriately */
    String detectFileFormat() {
        StringTokenizer st = new StringTokenizer(getConfig("FileInterfaces"));
        String ext = null, fmt = "???";
        int dotIndex;
        while (ext == null && st.hasMoreTokens()) {
            fmt = st.nextToken();
            ext = match(fileName, getConfig(fmt + ".sources"));
        }
        if (ext == null) { // if no match, go back & get 1st one
            st = new StringTokenizer(getConfig("FileInterfaces"));
            fmt = st.nextToken();
            ext = fileName.substring(fileName.length()-1); // last char in name
        }
        dotIndex = fileName.lastIndexOf(ext);
        if (!ext.startsWith(".") && !ext.startsWith("-")) {
            dotIndex = fileName.lastIndexOf(".", dotIndex);
        }
        baseName = dotIndex > 0 ? fileName.substring(0,dotIndex) : fileName + "-out";
        return fmt;
    }

  /** Examine source file name to decide syntax */
    String detectSyntax() {
        StringTokenizer st = new StringTokenizer(getConfig("SyntaxInterfaces"));
        String ext = null, syn = "???";
        int dotIndex;
        while (ext == null && st.hasMoreTokens()) {
            syn = st.nextToken();
            ext = match(fileName, getConfig(syn + ".sources"));
        }
        if (ext == null) { // if no match, go back & get 1st one
            st = new StringTokenizer(getConfig("SyntaxInterfaces"));
            syn = st.nextToken();
        }
        return syn;
    }


  /**
   * Main routine for weaving a file.  Call with the source filename.
   */
    public void weave (String fileName, String resultName, boolean weaving) {
        String lang;

        message((weaving ? "Weaving" : "Tangling")
            + " source file \"" + fileName + "\" ...");


        this.fileName = fileName;
        this.weaving = weaving;
        fileDir = (new File(fileName).getAbsolutePath());
        int dend = fileDir.lastIndexOf(File.separator);
        fileDir = fileDir.substring(0,dend);


      //---Nice predefined chunk for hiding code
        labeledChunks.put("hidden", "#1\n");

      //---set up engineMap
        StringTokenizer st = new StringTokenizer(getConfig("Languages"));
        while(st.hasMoreTokens()) {
            String lng = st.nextToken(),
                eng = getConfig(lng + ".engine");
            if (eng != null) engineMap.put(lng, eng);
            else engineMap.put(lng, lng); // engine same as lng if not found
        }

      //---set up file and syntax interfaces
        if (fmt == null) fmt = detectFileFormat();

      // -- set up system properties possibly needed by clients
        if (target == null) target = getConfig(fmt + ".target");
        config.setProperty("sys.target", target);
        config.setProperty("sys.basename", baseName);
        if (weaving)
            message("Result target type: " + target);

        try {
            filei = (FileInterface)(Class.forName(getConfig(fmt + ".class"))).newInstance();
            filei.setParent(this);
        }
        catch (Exception re) {
            re.printStackTrace(System.err);
            error("Unable to load needed file interface", Err.SYSERR);
        }

      // get ready to go...
        filei.readSourceFile(fileName);
        globalOpts.putAll(factoryOpts);   // start with factory options

      // First pass -- find tags in source file and organize the work ---
        Tag tag = null;
        for (tagNo=1, chunkNo = 0;  tag != Tag.noMore;  tagNo++)  {
            tag = filei.nextTag();
            if (tag == Tag.noMore) break;

            String sig = "!/swv-bkmrk" + tagNo + "/";

            if (tag == Tag.chunk) {  // Code chunk --------------------
                chunkNo++;
                lang = filei.getLang();
                String eng = engineMap.get(lang);
                if (eng == null) error(filei.getPosition() + ": No engine defined for " + lang, Err.SRCERR);
                EngineInterface ei  = engines.get(eng);

              // update the options
                currentOpts.clear();
                currentOpts.putAll(globalOpts);
                HashMap<String,String> lso = langSpecOpts.get(lang);
                if (lso != null) currentOpts.putAll(lso);
                parseOptions(filei.getOptions(), currentOpts);
// moved from later code...
                optHistory.put(sig, new HashMap<String,String>(currentOpts));

                boolean restart = isTrue("restart");

                if (restart || ei == null) { // need to instantiate a new engine
                    try {
                        String cls = getConfig(eng + ".class");
                        if (cls == null)
                            error("No engine is configured for " + eng, Err.NOKEY);
                        ei = (EngineInterface)(Class.forName(cls)).newInstance();
                        engines.put(eng, ei);
                        allEngines.add(ei);
                        ei.setParents(this, filei);
                        if (!ei.openCodeFile(restart ? baseName+chunkNo : baseName))
                            error("Can't open code file for " + eng, Err.IOERR);
                        ei.putStartup(weaving);
                        if (weaving) {
                            String bin = getConfig(eng + ".binary");
                            if (bin == null)
                                error("No binary is configured for " + eng, Err.NOBIN);
                            ei.setBinary(bin);
                        }
                        if (!restart) {   // check for existence of lang-specific opts
                            String lsospec = getConfig(lang + ".options");
                            if (lsospec != null) {
/* I don't think I needed to define a new lso map
                                HashMap<String,String> lshm = new HashMap<String,String>(4);
                                parseOptions(lsospec, lshm);
                                langSpecOpts.put(lang, lshm);
                                currentOpts.putAll(lshm);
REMOVE These when sure and use subst code below
*/
                                if (lso == null) lso = new HashMap<String,String>(4);
                                parseOptions(lsospec, lso);
                                langSpecOpts.put(lang, lso);
                                currentOpts.putAll(lso);
                              // Need to override the opt history for this chunk
                                optHistory.put(sig, new HashMap<String,String>(currentOpts));
/* send of subst code */
                            }
                        }
                    }
                    catch (Exception re) {
                        re.printStackTrace(System.err);
                        error("Unable to create needed interface(s)", Err.REFLECT);
                    }
                }

              // handle the code chunk.
                boolean eval = !dryrun && isTrue("eval"), fig = isTrue("fig"),
                    echo = isTrue("echo"), hide = isTrue("hide"),
                    savecode = isTrue("savecode"), saveout = isTrue("saveout"),
                    savefig = isTrue("savefig");

                // resolve implications
                if (getOption("results") != null) if (getOption("results").equals("hide")) hide = true;
                if (!eval) hide = true;
                if (savefig) fig = true;
                if (savecode) echo = true;
                if (saveout) hide = false;

                String savedCode = "",  codeListing = "",
                    label = getOption("label"), chunkName = baseName + "-",
                    prompt = getOption("prompt");

                if (label == null) {
                    label = "lastchunk";
                    chunkName += chunkNo + "-" + lang;
                }
                else
                    chunkName += label;
                if (eval) {
                    ei.putCode("");
                    ei.putComment("---- Chunk " + chunkNo + ": \"" + chunkName + "\" -----");
                    if (weaving) {
                        if (!hide) {
                            ei.putSeparator("!-StatWeave-" + sig + "out");
//                            optHistory.put(sig, new HashMap<String,String>(currentOpts));
                        }
                        if (fig) {
//                            if (hide)
//                                optHistory.put(sig, new HashMap<String,String>(currentOpts));
                            FigFile ff = ei.setupFig(chunkName);
                            allFigs.addElement(ff);
                            figFiles.put(sig + "fig", ff);
                        }
                    }
                }
                if (echo && (prompt == null))
                    prompt = lang + getOption("ompt", "> ");
                // get the code chunk
                for (String line = ""; line != null;  ) {
                    line = filei.getCodeLine();
                    if (line == null)
                        break;
                    if (line.startsWith("!coderef")) {
                        boolean showref = isTrue("showref");
                        String lab = line.substring(9);
                        if (lab.startsWith("*")) {   // show ref if label starts with *
                            showref = true;
                            lab = lab.substring(1);
                        }
                        String sc = labeledChunks.get(lab);
                        if (sc == null)
                            error(filei.getPosition() + ": Referencing unknown label \"" + lab +"\"", Err.SRCERR);
                        Vector<String> args = filei.getArgs();
                        for (int a=0; a<args.size(); a++) {
                            String arg = args.elementAt(a);
                            sc = sc.replaceAll("#"+(a+1), arg);
                        }
                        savedCode += sc;
                        if (eval) ei.putCode(sc);
                        if (showref && echo) {
                            sc += "\n";
                            int pos0, pos1;
                            for (pos0 = 0;  (pos1 = sc.indexOf("\n", pos0)) > 0; pos0=pos1+1) {
                                codeListing += prompt + sc.substring(pos0, pos1+1);
                            }
                        }
                    }
                    else if (line != null) {
                        savedCode += line + "\n";
                        if (echo) codeListing += prompt + line + "\n";
                        if (eval) ei.putCode(line);
                    }
                }

                // put together the bookmarks we need
                String bkmrk = "";
                if (echo && !savecode)
                    bkmrk += sig + "code";
                if (eval && weaving && !hide) {
                    ei.putSeparator("@-StatWeave-" + sig + "out");
                    if (!saveout) bkmrk += sig + "out";
                }
                if (weaving && eval && fig) {
                    ei.closeFig();
                    if (!savefig) bkmrk += sig + "fig";
                }
                if (weaving) filei.saveBookmark(bkmrk);

                if (echo)
                    if (savecode) { // save stuff for later
                        recallCode.put(label, codeListing);
                        recallCodeSig.put(label, sig);
                    }
                    else // put it now
                        if (weaving) filei.replaceBookmark(sig+"code", codeListing, Tag.recallCode);

                if (saveout && weaving)
                    recallOutSig.put(label, sig+"out");

                if (savefig && weaving) {
                    recallFigSig.put(label, sig+"fig");
                }

                // strip off the last "\n" before saving it
                labeledChunks.put(label, savedCode.substring(0, savedCode.length()-1));


            } //---if(tag==Tag.chunk)

            else if (tag == Tag.expr) {
                if (weaving) {
                    String extsig = sig + recIndex++ + "expr",
                        lng = filei.getLang(), expr = filei.getExpr(),
                        eng = engineMap.get(lng);
                    if (eng == null)
                        error(filei.getPosition() + ": No engine defined for " + lng, Err.SRCERR);
                    EngineInterface ei  = engines.get(eng);
                    if (ei == null)
                        error(filei.getPosition() + ": No engine yet active for " + lng, Err.SRCERR);
                    if (!dryrun) {
                        ei.putSeparator("!-StatWeave-" + extsig);
                        ei.putExpr(expr);
                        ei.putSeparator("@-StatWeave-" + extsig);
                        filei.saveBookmark(extsig);
                    }
                    else
                        filei.saveBookmark("[expression]");
                }
            }

            else if (weaving && tag == Tag.recallCode) {
                String lbl = filei.getLabel(),
                    code = recallCode.get(lbl),
                    chunksig = recallCodeSig.get(lbl);
                currentOpts.clear();
                currentOpts.putAll(optHistory.get(chunksig));
                filei.saveBookmark(sig);
                filei.replaceBookmark(sig, code, tag);
            }

            else if (weaving && tag == Tag.recallOutput) {
                String lbl = filei.getLabel(),
                    extsig = dryrun ? "[recall output]" : recallOutSig.get(lbl);
                filei.saveBookmark(extsig);
            }

            else if (weaving && tag == Tag.recallFig) {
                String lbl = filei.getLabel(),
                    extsig = dryrun ? "[recall figure]" : recallFigSig.get(lbl);
                filei.saveBookmark(extsig);
            }

          // Language-specific options ----------------
            else if (tag == Tag.langOpts) {
                lang = filei.getLang();
                HashMap<String,String> hm = langSpecOpts.get(lang);
                if (hm == null) {
                    hm = new HashMap<String,String>();
                    langSpecOpts.put(lang, hm);
                }
                parseOptions (filei.getOptions(), hm);
            }

          // Global options --------------------
            else if (tag == Tag.globalOpts) {
                parseOptions (filei.getOptions(), globalOpts);

              // -- new language definition?
                String nl = globalOpts.get("newlang");
                if (nl != null) try {
                    int colpos = nl.indexOf(":");
                    if (colpos < 0)
                        throw new IllegalArgumentException("Syntax error in \"newlang = " + nl + "\" -- ignored");
                    String ol = nl.substring(colpos+1), nle;
                    nl = nl.substring(0,colpos);
                    nle = engineMap.get(ol);  // guarantee we're mapping to an engine
                    if (nle == null)
                        throw new IllegalArgumentException("No engine or language named \'" + ol + "\"");
                    engineMap.put(nl, nle);
                }
                catch (IllegalArgumentException iae) {
                    warning(iae.getMessage());
                }
                globalOpts.remove("newlang"); // this isn't a real option
            }

        } // --- end of 1st pass

        if (allEngines.size() == 0) {
            warning("Note: No code chunks were found!");
        }
        // close all the code files
        for (Iterator<EngineInterface> i=allEngines.iterator(); i.hasNext(); ) {
            i.next().closeCodeFile();
        }
        if (!weaving) {
            message("The following code files have been created:");
            for (Iterator<EngineInterface> i=allEngines.iterator(); i.hasNext(); )
                message("\t"+ i.next().getFileNames()[0]);
            return;  // we're done tangling!
        }

// -------- Post-processing ---------------

      // First, we need to run all the code files in sequence
        try {
            if (!dryrun) for (Iterator<EngineInterface> i=allEngines.iterator(); i.hasNext(); ) {
                Process prc = i.next().runCode();
                if (prc != null) {
                    try {
                        BufferedReader prcOut = new BufferedReader(
                            new InputStreamReader(prc.getInputStream()));
                        for (String prcLine = prcOut.readLine(); prcLine != null; prcLine = prcOut.readLine())
                            message(prcLine);
                        }
                    catch (IOException ioe) {
                        error("Error in reading process output", Err.IOERR);
                    }
                    int status = prc.waitFor();
                    if (status > 0)
                        error("Runtime error " + status + " encountered in client programs", Err.CHILDERR);
                }
            }
            else
                message("*** Dry Run mode: No code is evaluated ***");
        }
        catch (InterruptedException e) {
            error("Client process was interrupted", Err.CHILDINT);
        }


    // Go through each output file .....
        if (!dryrun) for (Iterator<EngineInterface> i=allEngines.iterator(); i.hasNext(); ) {
            String fn = i.next().getFileNames()[1];
            BufferedReader inStream;
            try {
                File file = new File(fn);
                inStream = new BufferedReader(new FileReader(file));
                for (String line = inStream.readLine(); line != null; line = inStream.readLine()) {
                    int pos = line.indexOf("!-StatWeave-");
                    if (pos >= 0) {
                        String bkmrk = line.substring(pos + 12), outChunk = "",
                            sig = bkmrk.endsWith("out") ? bkmrk.substring(0,bkmrk.length()-3) : bkmrk;
                        if (bkmrk.endsWith("expr")) {
                            for (line = inStream.readLine(); line.indexOf("@-StatWeave-") < 0; line = inStream.readLine())
                                outChunk += line + " ";
                            filei.replaceBookmark(bkmrk, outChunk.trim(), Tag.expr);
                        }
                        else {
                            currentOpts.clear();
                            currentOpts.putAll(optHistory.get(sig));
                            for (line = inStream.readLine(); line.indexOf("@-StatWeave-") < 0; line = inStream.readLine())
                                outChunk += line + "\n";
                            filei.replaceBookmark(bkmrk, cleanUp(outChunk), Tag.recallOutput);
                        }
                    }
                }
                inStream.close();
                if (mopup >= 1) // delete intermediate file
                    file.delete();
            }
            catch(FileNotFoundException e) {
                error("Output file \"" + fn + "\" not found", Err.FILENF);
            }
            catch (IOException ioe) {
                ioe.printStackTrace();
                error("Unrecoverable error reading " + fn, Err.IOERR);
            } // end (lang!=null)
        } // -- end iterator for engine outputs

      // Embed the figures ...
        Iterator<String> fkeys = figFiles.keySet().iterator();
        while(fkeys.hasNext()) { // Get bookmarks -- they are of form "SIGfig"
            String bkmrk = fkeys.next(), sig = bkmrk.substring(0, bkmrk.length()-3);
            currentOpts.clear();
            currentOpts.putAll(optHistory.get(sig));
            FigFile thefig = figFiles.get(bkmrk);
            if (thefig != null)
                filei.embedPlots(bkmrk, thefig, new int[] {1}); // Need to soup this up later
            else
                filei.replaceBookmark(bkmrk,
                    "[Figure not generated due to errors]\n",
                    Tag.recallOutput);
        }

        filei.writeResults(mopup);

      // Delete all code files if appropriate
        if (mopup >= 1) {
            for (Iterator<EngineInterface> i = allEngines.iterator(); i.hasNext(); ) {
                i.next().deleteCodeFile();
            }
        }

      // Delete all figure files if appropriate
        if (mopup >= 2) {
            for (Iterator<FigFile> i = allFigs.iterator(); i.hasNext(); ) {
                FigFile ff = i.next();
                File file = new File(ff.fileName);
                file.delete();
            }
        }


    } // --- end of weave() ---



  /** Main routine for weaving a file.  Call with the source filename. */
    public void weave (String fileName) {
        weave(fileName, null, true);
    }

  /** Main routine for tangling a file.  Call with the source filename. */
    public void tangle (String fileName) {
        weave(fileName, null, false);
    }

  // Helper for parsing flags
    static String chkArg(String args[], int i) {
        return (i < args.length ? args[i] : null);
    }

  /** Command-line syntax message */
    static void cmdLineSyntax() {
        System.out.println("\nSyntax: statweave [options] sourceFileName [resultFileName]");
        System.out.println("\nOptions:");
        System.out.println("\t--weave        Weave into result file");
        System.out.println("\t--tangle       Extract the embedded code for each engine");
        System.out.println("\t--cleanup      Delete intermediate files");
        System.out.println("\t--tidyup       Delete non-graphics intermediate files");
        System.out.println("\t--keepall      Do not delete intermediate files");
        System.out.println("\t--config FILE  Use FILE as configuration file");
        System.out.println("\t--target EXT   Make output file of type EXT");
        System.out.println("\t--dryrun       Do not evaluate any code chunks");
    }


  /** Command-line interface */
    public static void main (String args[]) {
        boolean weaving = true, cleanup = false, tidyup = true, dryrun=false;
        String fileName=null, resultName = null, cfgFile = null, target=null;
        if (args.length == 0 ) {
            cmdLineSyntax();
            System.exit(0);
        }
        // Go through command-line options
        for (int i=0; i < args.length; i++) {
            if (args[i].startsWith("-")) {
                if (args[i].equalsIgnoreCase("--weave"))   weaving = true;
                else if (args[i].equalsIgnoreCase("--tangle"))   weaving = false;
                else if (args[i].equalsIgnoreCase("--cleanup"))   cleanup = true;
                else if (args[i].equalsIgnoreCase("--tidyup"))   tidyup = true;
                else if (args[i].equalsIgnoreCase("--keepall"))  cleanup = tidyup = false;
                else if (args[i].equalsIgnoreCase("--config"))  cfgFile = chkArg(args, ++i);
                else if (args[i].equalsIgnoreCase("--target"))  target = chkArg(args, ++i).toLowerCase();
                else if (args[i].equalsIgnoreCase("--dryrun"))  dryrun = true;
                else System.out.println("Warning: option \"" + args[i] + "\" is not available and was ignored.");
            }
            else {
                if (fileName == null) fileName = args[i];
                else resultName = args[i];
            }
        }
        StatWeave swv = new StatWeave(cfgFile);
        swv.mopup = cleanup ? 2  : tidyup ? 1   : 0;
        swv.target = target;
        swv.dryrun = dryrun;
        swv.weave(fileName, resultName, weaving);
    }

}

